"""
Invoke tasks by enqueing them in the datastore. Not much to do here :-)
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from frocket.invoker.impl.async_invoker import AsyncInvoker


class WorkQueueInvoker(AsyncInvoker):
    def _enqueue(self, requests) -> None:
        self._datastore.enqueue(requests)
